/*
* Copyright (c) 2006-2009 Erin Catto http://www.box2d.org
*
* This software is provided 'as-is', without any express or implied
* warranty.  In no event will the authors be held liable for any damages
* arising from the use of this software.
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution.
*/

#ifndef ROBOTARM_H
#define ROBOTARM_H

#include <vector>
#include <set>
#include <map>
#include "../Framework/b2dJson.h"

#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f

#define ITERATE_SET(theType, theSet) \
for (std::set<theType>::iterator it = theSet.begin(); it != theSet.end(); ++it)

    struct armpos {
    float shoulderAngle;
    float elbowAngle;
    float wristAngle;
};

class RobotArm : public Test
{
public:

    RobotArm()
    {
        m_prodSensorCount = 0;

        //load world from JSON file
        b2dJson json;
        b2World* world = json.readFromFile("robot.json");
        if ( world ) {
            //replace testbed's world
            delete m_world;
            m_world = world;

            //re-set standard testbed stuff
            m_world->SetContactListener(this);
            m_world->SetDebugDraw(&m_debugDraw);
            b2BodyDef bodyDef;
            m_groundBody = m_world->CreateBody(&bodyDef);//needed for testbed mousejoint

            m_shoulderJoint = (b2RevoluteJoint*)json.getJointByName("shoulder");
            m_elbowJoint = (b2RevoluteJoint*)json.getJointByName("elbow");
            m_wristJoint = (b2RevoluteJoint*)json.getJointByName("wrist");
            m_handJoint = (b2PrismaticJoint*)json.getJointByName("hand");
            m_prodJoint = (b2PrismaticJoint*)json.getJointByName("prodjoint");
            m_prodSensor = json.getFixtureByName("prodsensor");
            m_prodFixture = json.getFixtureByName("prodfixture");

            vector<b2Body*> boxes;
            json.getBodiesByName("box", boxes);
            for (int i = 0; i < boxes.size(); i++)
                boxes[i]->SetSleepingAllowed(false);

            loadArmPosition(1, "robot1.json");
            loadArmPosition(2, "robot2.json");
            loadArmPosition(3, "robot3.json");
            loadArmPosition(4, "robot4.json");
            loadArmPosition(5, "robot5.json");
            loadArmPosition(6, "robot6.json");
            loadArmPosition(7, "robot7.json");
            loadArmPosition(8, "robot8.json");
            loadArmPosition(9, "robot9.json");

            m_sequence.push_back(6);
            m_sequence.push_back(7);
            m_sequence.push_back(-1);
            m_sequence.push_back(6);
            m_sequence.push_back(4);
            m_sequence.push_back(8);
            m_sequence.push_back(9);
            m_sequence.push_back(-1);
            m_sequence.push_back(8);
            m_sequence.push_back(4);
            m_sequence.push_back(6);

            m_sequence.push_back(6);
            m_sequence.push_back(7);
            m_sequence.push_back(-1);
            m_sequence.push_back(6);
            m_sequence.push_back(3);
            m_sequence.push_back(-1);

            m_sequence.push_back(6);
            m_sequence.push_back(7);
            m_sequence.push_back(-1);
            m_sequence.push_back(6);
            m_sequence.push_back(4);
            m_sequence.push_back(5);
            m_sequence.push_back(-1);
            m_sequence.push_back(4);

            m_sequenceInd = 0;

            m_targetPosIndex = m_sequence[m_sequenceInd];
            m_handClosed = false;
            m_pauseSteps = 0;
        }
    }

    void loadArmPosition(int posIndex, string filename)
    {
        b2dJson json2;
        b2World* world2 = json2.readFromFile(filename.c_str());
        if ( world2 )
        {
            b2RevoluteJoint* shoulderJoint = (b2RevoluteJoint*)json2.getJointByName("shoulder");
            b2RevoluteJoint* elbowJoint = (b2RevoluteJoint*)json2.getJointByName("elbow");
            b2RevoluteJoint* wristJoint = (b2RevoluteJoint*)json2.getJointByName("wrist");
            armpos ap;
            ap.shoulderAngle = shoulderJoint->GetJointAngle();
            ap.elbowAngle = elbowJoint->GetJointAngle();
            ap.wristAngle = wristJoint->GetJointAngle();
            m_armPositions[posIndex] = ap;
            delete world2;
        }
    }

    void Keyboard(unsigned char key)
    {
        switch (key)
        {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            m_targetPosIndex = key-48;
            break;
        case 'c':
            m_handClosed = !m_handClosed;
            break;
        }
    }

    void KeyboardUp(unsigned char key)
    {
    }

    void SetMotorSpeeds(armpos& targetPos)
    {
        float shoulderDelta = targetPos.shoulderAngle - m_shoulderJoint->GetJointAngle();
        float elbowDelta = targetPos.elbowAngle - m_elbowJoint->GetJointAngle();
        float wristDelta = targetPos.wristAngle - m_wristJoint->GetJointAngle();

        shoulderDelta *= RADTODEG;
        elbowDelta *= RADTODEG;
        wristDelta *= RADTODEG;

        float angleCutoff = 1;
        float minAngleSpeed = 5;
        float maxAngleSpeed = m_handClosed ? 40 : 240;

        shoulderDelta = b2Clamp(shoulderDelta, -maxAngleSpeed, maxAngleSpeed);
        elbowDelta = b2Clamp(elbowDelta, -maxAngleSpeed, maxAngleSpeed);
        wristDelta = b2Clamp(wristDelta, -maxAngleSpeed, maxAngleSpeed);

        shoulderDelta *= DEGTORAD;
        elbowDelta *= DEGTORAD;
        wristDelta *= DEGTORAD;

        m_shoulderJoint->SetMotorSpeed(shoulderDelta);
        m_elbowJoint->SetMotorSpeed(elbowDelta);
        m_wristJoint->SetMotorSpeed(wristDelta);

    }

    bool GetArrived(armpos& targetPos)
    {
        float shoulderDelta = targetPos.shoulderAngle - m_shoulderJoint->GetJointAngle();
        float elbowDelta = targetPos.elbowAngle - m_elbowJoint->GetJointAngle();
        float wristDelta = targetPos.wristAngle - m_wristJoint->GetJointAngle();

        shoulderDelta *= RADTODEG;
        elbowDelta *= RADTODEG;
        wristDelta *= RADTODEG;

        float lim = (m_targetPosIndex == 8 || m_targetPosIndex == 4) ? 15 : 0.5f;
        return fabs(shoulderDelta) < lim && fabs(elbowDelta) < lim && fabs(wristDelta) < lim;
    }

    void Step(Settings* settings)
    {
        if ( m_pauseSteps > 0)
            m_pauseSteps--;
        if ( m_pauseSteps <= 0 && m_targetPosIndex )
        {
            armpos ap = m_armPositions[m_targetPosIndex];
            if ( !GetArrived(ap) )
            {                
                SetMotorSpeeds(ap);
            }
            else {
                if ( m_targetPosIndex == 3 || m_targetPosIndex == 5 || m_targetPosIndex == 9 )
                    m_handClosed = false;
                else if ( m_targetPosIndex == 7 )
                    m_handClosed = true;

                m_sequenceInd++;
                if ( m_sequenceInd >= m_sequence.size() )
                    m_sequenceInd = 0;
                m_targetPosIndex = m_sequence[m_sequenceInd];
                if ( m_targetPosIndex == -1 )
                {
                    m_pauseSteps = 45;
                    m_sequenceInd++;
                    if ( m_sequenceInd >= m_sequence.size() )
                        m_sequenceInd = 0;
                    m_targetPosIndex = m_sequence[m_sequenceInd];
                }
            }
        }

        if ( m_handClosed )
            m_handJoint->SetMotorSpeed(1);
        else
            m_handJoint->SetMotorSpeed(-1);

        if ( m_prodSensorCount > 0 )
            m_prodJoint->SetMotorSpeed(0.125);
        else
            m_prodJoint->SetMotorSpeed(-0.125);

        Test::Step(settings);

        m_debugDraw.DrawString(5, m_textLine, "Prod sensor count: %d", m_prodSensorCount);
        m_textLine += 15;
    }

    void PreSolve(b2Contact* contact, const b2Manifold* oldManifold)
    {
    }

    void BeginContact(b2Contact* contact)
    {
        if ( contact->GetFixtureA() == m_prodSensor || contact->GetFixtureB() == m_prodSensor )
            m_prodSensorCount++;
    }

    void EndContact(b2Contact* contact)
    {
        if ( contact->GetFixtureA() == m_prodSensor || contact->GetFixtureB() == m_prodSensor )
            m_prodSensorCount--;
    }

    void PostSolve(b2Contact* contact, const b2ContactImpulse* impulse)
    {
    }

    static Test* Create()
    {
        return new RobotArm();
    }

    b2RevoluteJoint* m_shoulderJoint;
    b2RevoluteJoint* m_elbowJoint;
    b2RevoluteJoint* m_wristJoint;
    b2PrismaticJoint* m_handJoint;
    b2PrismaticJoint* m_prodJoint;
    b2Fixture* m_prodSensor;
    b2Fixture* m_prodFixture;
    int m_prodSensorCount;

    map< int, armpos > m_armPositions;
    int m_targetPosIndex;
    bool m_handClosed;

    vector<int> m_sequence;
    int m_sequenceInd;
    int m_pauseSteps;

};

#endif


